'use strict';
let Validator = require('ajv');
const csType_DlmProfile = '.dwfdlmprofile';
let ajv = new Validator({ verbose: true, allErrors: true, extendRefs: true });
ajv.addSchema(require('./schema/dlmprofile.schema.json'), csType_DlmProfile);
function getSeverity() {
  return {
    warning: 'warning',
    error: 'error',
    info: 'info'
  };
}
let errorsEnum = {
  'TYPE1': 'Cannot find column type.',
  'TYPE2': 'column value does not match column type, ',
  'TYPE3': 'None of key has value. At least one key has to have a value'
};
/**
 * To validate given schema symantic check
 *
 * @param {*} isSchemaType
 * @param {*} ioJSON
 *
 * @return obj  {
    bValid: Type boolean,
    oErrors: Type obj,
    sMsg: type array of object :: obj = error Message
  }
 */
let fValidateSchema = function (isSchemaType, ioJSON) {
  if (typeof (ioJSON) === 'string') {
    ioJSON = JSON.parse(ioJSON);
  }
  let severity = getSeverity();
  let retval = ajv.validate(isSchemaType, ioJSON);
  let aoMessages = [];
  if (ajv && ajv.errors) {
    ajv.errors.forEach(function (err) {
      if (err.params) {
        let lsMsg = `${err.dataPath}  ${err.message}  `;
        Object.keys(err.params).forEach(function (k) {
          lsMsg = ` ${lsMsg},  ${k}  = ` + err.params[k];
        });
        //msg.push({ 'msg': JSON.stringify(err.params) + '  ' + JSON.stringify(err.schemaPath) + ' ', severity: severity.error });
        aoMessages.push({ 'msg': lsMsg, severity: severity.error });
      }
    });
  }
  if (!retval) {
    aoMessages.unshift({ msg: isSchemaType + ' schema validation failed', severity: severity.error });
  }
  return {
    bValid: retval,
    oErrors: ajv.errors,
    aoMessages: aoMessages
  };
};
/**
 *
 * @param {*} iaNodes nodes to be checked/validated
 * @param {*} isAction coulld be key or value
 * @param {*} ioGlobalNodes  nodes against whcih we validate iaNodes
 * @param {*} isMsg node name os iaNodes in dlmProfile
 *
 * @return [{}, {}] while objects are errors
 */
let fValidate = function (iaNodes, isAction, ioGlobalNodes, isMsg) {
  let faultyContents = [];
  let foundFlg = false;
  iaNodes.forEach(function (node) {
    Object.keys(ioGlobalNodes).forEach(function (key) {
      if ((isAction === 'value') ? ioGlobalNodes[key] === node : key === node) {
        foundFlg = true;
        //break ioGlobalNodes;
      }
    });
    if (!foundFlg) {
      let lObject = {};
      lObject['nodeName'] = isMsg;
      lObject['node'] = node;
      faultyContents.push(lObject);
    }
    foundFlg = false;
  });
  return faultyContents;
};
/**
 *
 * @param {*} ioDlmProfile
 * ioDLMProfile has all nodes from profile which needs to be validated for contents
 * this methods calls another function fValidate to compare contents of two particular nodes
 * once all nodes are checked, this method compiles all error in a msg and return an object with results
 */
let fSchemaValidationHelper = function (ioDlmProfile) {
  let laContentErrors = [];
  let conns = ioDlmProfile.globalConnections;
  //loops on conn node and gets source and target nodes to be validated agains global nodes
  Object.keys(conns).forEach(function (key) {
    let tempConn = conns[key];
    tempConn.shift(); // ['c1', 'n1', 'n2'] --> ['n1', 'n2']
    laContentErrors = laContentErrors.concat(fValidate(tempConn, 'value', ioDlmProfile.globalNodes, 'conn'));
  });

  //laContentErrors = laContentErrors.concat(fValidate(ioDlmProfile.uiNodes, 'key', ioDlmProfile.globalNodes, 'ui'));
  laContentErrors = laContentErrors.concat(fValidate(ioDlmProfile.designNodes, 'key', ioDlmProfile.globalNodes, 'design'));
  laContentErrors = laContentErrors.concat(fValidate(ioDlmProfile.runtimeNodes, 'value', ioDlmProfile.globalNodes, 'runtime'));
  laContentErrors = laContentErrors.concat(fValidate(ioDlmProfile.columns2Check, 'key', ioDlmProfile.columnsAvailable, 'Source <-> Target'));
  laContentErrors = laContentErrors.concat(validateColumnTypes(ioDlmProfile.columnTypes2Check, ioDlmProfile.columnsAvailableWithPath));
  Object.keys(ioDlmProfile.keys2Check).forEach(function (k) {
    let latemp = [];
    if (!handleKeyContentWithMessage('RESULT', ioDlmProfile.keys2Check, null, k)) {
      latemp.push(getErrorMessage(k + '  :', errorsEnum.TYPE3));
    }
    laContentErrors = laContentErrors.concat(latemp);
  });

  let objDesign = {};
  //transfarming from object to array
  (ioDlmProfile.designNodes).forEach(function (item, i) {
    objDesign[i] = item;
  });
  laContentErrors = laContentErrors.concat(fValidate(Object.keys(ioDlmProfile.globalNodes), 'value', objDesign, 'global'));
  let aoMessages = [];// (ioDlmProfile.id && ioDlmProfile.namespace) ? `Namespace : ${ioDlmProfile.namespace},  File :  ${ioDlmProfile.id}  has following errors \n` : '';
  let retval = true;
  laContentErrors.forEach(function (fNode) {
    retval = false;
    if (fNode.errorType) {
      aoMessages.push(fNode.message);
    } else {
      let lsMissingFrom = (fNode.nodeName === 'global') ? ' Design' : 'Global';
      aoMessages.push({ 'msg': `'${fNode.nodeName}' contains  object ${fNode.node} which does not exist  in  ${lsMissingFrom} \n`, 'severity': getSeverity().error });
    }
  });
  if (!retval) {
    aoMessages.unshift({ msg: ' schema content validation failed', 'severity': getSeverity().error });
  }
  return {
    bValid: retval,
    aoMessages: aoMessages
  };
};
/**
 *
 * @param {*} objColumn
 * @param {*} loaColumnTypes2Check
 */
function extractDataTypesForContentCheck(objColumn, loaColumnTypes2Check) {
  if (loaColumnTypes2Check.hasOwnProperty(objColumn.column)) {
    let tempObj = loaColumnTypes2Check[objColumn.column];
    if (objColumn.lowValue) {
      if (tempObj.lowValue) {
        let tempLowArray = tempObj.lowValue;
        tempLowArray.push(objColumn.lowValue);
        tempObj.lowValue = tempLowArray;
      } else {
        let tempLowArray = [];
        tempLowArray.push(objColumn.lowValue);
        tempObj.lowValue = tempLowArray;
      }
    }
    if (objColumn.highValue) {
      if (tempObj.highValue) {
        let tempHighValueArray = tempObj.highValue;
        tempHighValueArray.push(objColumn.highValue);
        tempObj.highValue = tempHighValueArray;
      } else {
        let tempHighValueArray = [];
        tempHighValueArray.push(objColumn.highValue);
        tempObj.highValue = tempHighValueArray;
      }
    }
    loaColumnTypes2Check[objColumn.column] = tempObj;
  } else {
    let tempLowArray = [];
    let tempHighValueArray = [];
    if (objColumn.lowValue) {
      tempLowArray.push(objColumn.lowValue);
    }
    if (objColumn.highValue) {
      tempHighValueArray.push(objColumn.highValue);
    }
    let obj = {};
    obj.lowValue = tempLowArray;
    obj.highValue = tempHighValueArray;
    loaColumnTypes2Check[objColumn.column] = obj;
  }
}
function valuesChecks(laValues, ioTypeAndPath, iaFaultyContents) {
  laValues.forEach(function (val) {
    if (!parseSQLType(val, ioTypeAndPath.type)) {
      iaFaultyContents.push(getErrorMessage(errorsEnum.TYPE2, `Value: ${val}, Type : ${ioTypeAndPath.type}, Path: ${ioTypeAndPath.path}`));
    }
  });
}
function validateColumnTypes(ioaColumnTypes2Check, ioColumnsAvailableWithPath) {
  let faultyContents = [];
  Object.keys(ioaColumnTypes2Check).forEach(function (key) {
    if (ioColumnsAvailableWithPath.hasOwnProperty(key) && ioColumnsAvailableWithPath[key]) {
      let loTypeAndPath = ioColumnsAvailableWithPath[key];
      if (ioaColumnTypes2Check[key].highValue.length > 0) {
        valuesChecks(ioaColumnTypes2Check[key].highValue, loTypeAndPath, faultyContents);
      }
      if (ioaColumnTypes2Check[key].lowValue.length > 0) {
        valuesChecks(ioaColumnTypes2Check[key].lowValue, loTypeAndPath, faultyContents);
      }
    } else {
      faultyContents.push(getErrorMessage(errorsEnum.TYPE1, ' : ' + key));
    }
  });
  return faultyContents;
}

function getErrorMessage(errorType, additionalMessage) {
  let oErrorMessage = {};
  oErrorMessage['errorType'] = errorType;
  oErrorMessage['message'] = { 'msg': errorType + '  ' + additionalMessage, severity: getSeverity().error };
  return oErrorMessage;
}
/**
 *
 * @param {*} sValue
 * @param {*} sType
 */
let parseSQLType = function (sValue, sType) {
  var ERR = false;
  // SOURCE: https://help.sap.com/viewer/4fe29514fd584807ac9f2a04f6754767/2.0.01/en-US/4ee2f261e9c44003807d08ccc2e249ac.html
  var fnIntegerCheck = function (nr, bytes) {
    if (/[^0-9]/g.exec(nr)) { return ERR; }
    nr = Number.parseInt(nr);
    if (Number.isNaN(nr)) { return ERR; }

    var neg = bytes > 1;
    var bits = bytes * 8;
    var min = neg ? -Math.pow(2, bits) / 2 : 0;
    var max = neg ? Math.pow(2, bits) / 2 - 1 : Math.pow(2, bits) - 1;
    if (nr < min || nr > max) { return ERR; }
    return true;
  };
  var fnDecimalCheck = function (nr/*, type*/) {
    nr = Number.parseFloat(nr);
    if (Number.isNaN(nr)) { return ERR; }
    if (/[^0-9]/g.exec(nr)) { return ERR; }

    // var dec = type.toUpperCase().indexOf('DEC') > -1;
    // TODO: Intorduce size checks
    return true;
  };

  var fnEscapeString = function (str, size) {
    // TODO: Introduce size checks
    if (size > 0 && str.length > size) { return ERR; }
    return true;
  };

  var fnBooleanCheck = function (val) {
    var trueValues = ['TRUE', 'true', '1'];
    var falseValues = ['FALSE', 'false', '0'];
    if (trueValues.indexOf(val) > -1) {
      return true;
    }
    if (falseValues.indexOf(val) > -1) {
      return true;
    }
    return ERR;
  };

  sType = sType || '';
  switch (sType.split('(')[0].toUpperCase()) {
    // Numeric types
    case 'TINYINT':
      return fnIntegerCheck(sValue, 1);
    case 'SMALLINT':
      return fnIntegerCheck(sValue, 2);
    case 'INTEGER':
      return fnIntegerCheck(sValue, 4);
    case 'BIGINT':
      return fnIntegerCheck(sValue, 8);
    case 'SMALLDECIMAL':
      return fnDecimalCheck(sValue, sType);
    case 'DECIMAL':
      return fnDecimalCheck(sValue, sType);
    case 'REAL':
      return fnDecimalCheck(sValue, sType);
    case 'DOUBLE':
      return fnDecimalCheck(sValue, sType);

    // Boolean type
    case 'BOOLEAN':
      return fnBooleanCheck(sValue);

    // Character string types
    case 'VARCHAR':
      return fnEscapeString(sValue);
    case 'NVARCHAR':
      return fnEscapeString(sValue);
    case 'ALPHANUM':
      return fnEscapeString(sValue);
    case 'SHORTTEXT':
      return fnEscapeString(sValue);

    // Binary types
    case 'VARBINARY':
      return fnEscapeString(sValue);

    // Large Object types
    // TODO
    case 'BLOB':
      return fnEscapeString(sValue);
    case 'CLOB':
      return fnEscapeString(sValue);
    case 'NCLOB':
      return fnEscapeString(sValue);
    case 'TEXT':
      return fnEscapeString(sValue);
    case 'BINTEXT':
      return fnEscapeString(sValue);

    // Datetime types
    // TODO: Ensure proper dates and within the date range
    // SOURCE: https://help.sap.com/viewer/4fe29514fd584807ac9f2a04f6754767/2.0.01/en-US/3f81ccc7e35d44cbbc595c7d552c202a.html
    case 'DATE':
      return fnEscapeString(sValue);
    case 'TIME':
      return fnEscapeString(sValue);
    case 'TIMESTAMP':
      return fnEscapeString(sValue);
    case 'SECONDDATE':
      return fnEscapeString(sValue);

    default:
      return sValue;
  }
};

/**
 *
 * @param {*} isAction
 * @param {*} iooKeys2Check
 * @param {*} isValue
 */
function handleKeyContentWithMessage(isAction, iooKeys2Check, isValue, isNodePath) {
  let lbResult = false;
  let currentObj = (iooKeys2Check.hasOwnProperty(isNodePath) ? iooKeys2Check[isNodePath] : {'emptyKey': 0, 'totalKeys': 0} );
  switch (isAction) {
    case ('INSERT'):
      if(isValue === '' || isValue === ' ' || isValue === '  ') {
        currentObj.emptyKey = ++currentObj.emptyKey;
        currentObj.totalKeys = ++currentObj.totalKeys;
      } else {
        currentObj.totalKeys = ++currentObj.totalKeys;
      }
      iooKeys2Check[isNodePath] = currentObj;
      break;
    case ('RESULT'):
      if (currentObj.emptyKey < currentObj.totalKeys) {
        lbResult = true;
      }
      return lbResult;
    default:
      return lbResult;
  }
}

/**
 * This function receive a json file object
 * returns an object with 3 key:value {valid: (true/false), while error and msg keys both have details about missing/wrong data in dlmprofile}
 * it checks following nodes
 * 'conn' nodes against Global nodes
 * 'ui' nodes against Global nodes
 * 'design' nodes against Global nodes
 * 'runtime' nodes agaisnt  Global nodes
 * 'global' nodes against Design
 *
 */
let fValidateContents = function (ioJSON) {
  if (typeof (ioJSON) === 'string') {
    ioJSON = JSON.parse(ioJSON);
  }
  let loDlmProfile = {};
  let loTemporary = {};
  let laTemporary = [];
  let loColumnsAvailable = {};
  let loColumnsAvailableWithPath = {}; // {key: {type: lsType, 'path': lsPath }}
  let laColumns2Check = [];
  let loaColumnTypes2Check = {};
  let looKeys2Check = {}; // {key: {'emptyKey': 0, 'totalKeys': 0}}

  if (ioJSON.dlmprofileId && ioJSON.namespace) {
    loDlmProfile.id = ioJSON.dlmprofileId;
    loDlmProfile.namespace = ioJSON.namespace;
  }
  // reading all global nodes from json object, transforming and saving it for future use e.g {'name'(0, 1, 2) : id(n1, n2, n3 ...)}
  if (ioJSON.global.node) {
    Object.keys(ioJSON.global.node).forEach(function (key) {
      loTemporary[key] = ioJSON.global.node[key].id;
      //this whole block is to get and save source_to_target &target_to_source nodes.
      let lsType = ioJSON.global.node[key].type;
      if (lsType === 'dataStore') {
        let laObjectColumn = (ioJSON.global.node[key][lsType].object[ioJSON.global.node[key][ioJSON.global.node[key].type].object.type].column) ?
          ioJSON.global.node[key][lsType].object[ioJSON.global.node[key][ioJSON.global.node[key].type].object.type].column : [];
        let lsNodePath = `global.node[${key}][${lsType}].object.${ioJSON.global.node[key][ioJSON.global.node[key].type].object.type}.column`;

        laObjectColumn.forEach(function (objColumn) {
          if (!loColumnsAvailable.hasOwnProperty(objColumn.name)) {
            loColumnsAvailable[objColumn.name] = objColumn.type;
            loColumnsAvailableWithPath[objColumn.name] = {'type': objColumn.type, 'path': lsNodePath};
          }
          handleKeyContentWithMessage('INSERT', looKeys2Check, objColumn.key, lsNodePath);
        });
      } else if (lsType === 'rule') {
        let laObjectColumnSource2Target = (ioJSON.global.node[key][ioJSON.global.node[key].type][ioJSON.global.node[key][ioJSON.global.node[key].type].type].source_to_target.rule_elements) ?
          ioJSON.global.node[key][ioJSON.global.node[key].type][ioJSON.global.node[key][ioJSON.global.node[key].type].type].source_to_target.rule_elements : [];
        laObjectColumnSource2Target.forEach(function (objColumn) {
          extractDataTypesForContentCheck(objColumn, loaColumnTypes2Check);
          laColumns2Check.push(objColumn.column);
        });
        let laObjectColumnTarget2Source = (ioJSON.global.node[key][ioJSON.global.node[key].type][ioJSON.global.node[key][ioJSON.global.node[key].type].type].target_to_source.rule_elements) ?
          ioJSON.global.node[key][ioJSON.global.node[key].type][ioJSON.global.node[key][ioJSON.global.node[key].type].type].target_to_source.rule_elements : [];
        laObjectColumnTarget2Source.forEach(function (objColumn) {
          extractDataTypesForContentCheck(objColumn, loaColumnTypes2Check);
          laColumns2Check.push(objColumn.column);
        });
      }
    });
  }


  loDlmProfile.globalNodes = loTemporary;
  loDlmProfile.columnsAvailable = loColumnsAvailable;
  loDlmProfile.columnsAvailableWithPath = loColumnsAvailableWithPath;
  loDlmProfile.columns2Check = laColumns2Check;
  loDlmProfile.columnTypes2Check = loaColumnTypes2Check;
  loDlmProfile.keys2Check = looKeys2Check;
  loTemporary = {};
  // reading all conn nodes from json object, transforming and saving it for future use :: example {['c1', 'n1', 'n3'], ...}
  Object.keys(ioJSON.global.conn).forEach(function (key) {
    laTemporary = [];
    laTemporary.push(ioJSON.global.conn[key].id);
    laTemporary.push(ioJSON.global.conn[key].source);
    laTemporary.push(ioJSON.global.conn[key].target);

    loTemporary[key] = laTemporary;
  });
  loDlmProfile.globalConnections = loTemporary;
  loTemporary = {};
  laTemporary = [];
  /*
  // reading all ui nodes from json object, transforming and saving it for future use :: example {['0', '1', '2']}
  Object.keys(ioFile.ui.node).forEach(function (key) {
    laTemporary.push(key);
  });
  loDlmProfile.uiNodes = laTemporary;*/
  loTemporary = {};
  laTemporary = [];
  // reading all design nodes from json object, transforming and saving it for future use :: example {[same like others nodes save id]}
  Object.keys(ioJSON.design.node).forEach(function (key) {
    laTemporary.push(key);
  });
  loDlmProfile.designNodes = laTemporary;
  loTemporary = {};
  laTemporary = [];
  // reading all runtime nodes from json object, transforming and saving it for future use :: example {['n1', 'n2'] same again}
  if (ioJSON.runtime && ioJSON.runtime.node) {
    (ioJSON.runtime.node).forEach(function (obj) {
      laTemporary.push(obj.id);
    });
  }

  loDlmProfile.runtimeNodes = laTemporary;
  return fSchemaValidationHelper(loDlmProfile);
};

module.exports = {
  csType_DlmProfile: csType_DlmProfile,
  fValidateSchema: fValidateSchema,
  fValidateContents: fValidateContents
};
